import numpy as np

"""
Give, two x,y curves this gives intersection points,
autor: Sukhbinder
5 April 2017


Based on: http://uk.mathworks.com/matlabcentral/fileexchange/11837-fast-and-robust-curve-intersections
"""


def _rect_inter_inner(x1, x2):
    n1 = x1.shape[0] - 1
    n2 = x2.shape[0] - 1
    X1 = np.c_[x1[:-1], x1[1:]]
    X2 = np.c_[x2[:-1], x2[1:]]
    S1 = np.tile(X1.min(axis=1), (n2, 1)).T
    S2 = np.tile(X2.max(axis=1), (n1, 1))
    S3 = np.tile(X1.max(axis=1), (n2, 1)).T
    S4 = np.tile(X2.min(axis=1), (n1, 1))
    return S1, S2, S3, S4


def _rectangle_intersection_(x1, x2):
    S1, S2, S3, S4 = _rect_inter_inner(x1, x2)

    C1 = np.less_equal(S1, S2)
    C2 = np.greater_equal(S3, S4)

    return C1, C2


def _rectangle_intersection_check_(x1, y1, x2, y2):
    C1, C2 = _rectangle_intersection_(x1, x2)
    C3, C4 = _rectangle_intersection_(y1, y2)

    ii, jj = np.nonzero(C1 & C2 & C3 & C4)
    return ii, jj


def intersection(x1, y1, x2, y2):
    """
    INTERSECTIONS Intersections of curves.
       Computes the (x,y) locations where two curves intersect.  The curves
       can be broken with NaNs or have vertical segments.

    usage:
    x,y=intersection(x1,y1,x2,y2)

        Example:
        a, b = 1, 2
        phi = np.linspace(3, 10, 100)
        x1 = a*phi - b*np.sin(phi)
        y1 = a - b*np.cos(phi)

        x2=phi
        y2=np.sin(phi)+2
        x,y=intersection(x1,y1,x2,y2)

        plt.plot(x1,y1,c='r')
        plt.plot(x2,y2,c='g')
        plt.plot(x,y,'*k')
        plt.show()

    """
    x1 = np.asarray(x1)
    x2 = np.asarray(x2)
    y1 = np.asarray(y1)
    y2 = np.asarray(y2)

    ii, jj = _rectangle_intersection_check_(x1, y1, x2, y2)
    n = len(ii)

    dxy1 = np.diff(np.c_[x1, y1], axis=0)
    dxy2 = np.diff(np.c_[x2, y2], axis=0)

    T = np.zeros((4, n))
    AA = np.zeros((4, 4, n))
    AA[0:2, 2, :] = -1
    AA[2:4, 3, :] = -1
    AA[0::2, 0, :] = dxy1[ii, :].T
    AA[1::2, 1, :] = dxy2[jj, :].T

    BB = np.zeros((4, n))
    BB[0, :] = -x1[ii].ravel()
    BB[1, :] = -x2[jj].ravel()
    BB[2, :] = -y1[ii].ravel()
    BB[3, :] = -y2[jj].ravel()

    for i in range(n):
        try:
            T[:, i] = np.linalg.solve(AA[:, :, i], BB[:, i])
        except:
            T[:, i] = np.inf

    in_range = (T[0, :] >= 0) & (T[1, :] >= 0) & (T[0, :] <= 1) & (T[1, :] <= 1)

    # get indexes of intersecting elements
    ij0 = [ii[in_range], jj[in_range]] + T[:2, in_range]
    ij0 = ij0.T
    # get coordinates of intersecting elements
    xy0 = T[2:, in_range]
    xy0 = xy0.T

    return xy0[:, 0], xy0[:, 1], ij0[:, 0], ij0[:, 1]


if __name__ == "__main__":
    import matplotlib.pyplot as plt

    # a piece of a prolate cycloid, and am going to find
    a, b = 1, 2
    phi = np.linspace(3, 10, 100)
    x1 = a * phi - b * np.sin(phi)
    y1 = a - b * np.cos(phi)

    x2 = phi
    y2 = np.sin(phi) + 2
    x, y, i, j = intersection(x1, y1, x2, y2)
    print(f"curves intersect at elements {i, j}, corresponding to coordinates {x,y}")
    plt.plot(x1, y1, c="r")
    plt.plot(x2, y2, c="g")
    plt.plot(x, y, "*k")
    plt.show()
